@@ -19,17 +19,18 @@ import json
 import traceback
 import urllib
 import warnings
-from typing import List, Optional, cast
+from typing import List, Optional, cast, Union
 
 import deprecation
 import pydash
-from cortex.serviceconnector import _Client
-from cortex.utils import get_logger
+from cortex.client import Client as CortexClient
+from cortex.serviceconnector import _Client, ServiceConnector
 from requests.exceptions import HTTPError
 
 from cortex_common.types import MessageResponse, Profile, ProfileSchema, EntityEvent, ProfileSummary, \
-    ProfileAttributeType, ProfileVersionSummary, \
+    ProfileAttributeType, ProfileVersionSummary, EntityRelationshipEvent, \
     ProfileAttributeSummary, HistoricProfile, HistoricProfileAttribute
+from cortex_common.utils import get_logger
 from cortex_common.utils import head, dicts_to_classes, dict_to_attr_class, fold_list
 from cortex_profiles.build.attributes.utils.etl_utils import turn_attribute_into_entity_event
 
@@ -53,11 +54,22 @@ class ProfilesRestClient(_Client):
         'schemas':  'graph/profiles/schemas',
         'schema':   'graph/profiles/schemas/{schemaId}',
         'profile':  'graph/profiles/{profileId}',
+        'profile-cache': 'graph/profile-cache',
     }
 
+    @staticmethod
+    def from_cortex_client(cortex_client:CortexClient):
+        return ProfilesRestClient(cortex_client, version=3)
+
     @classmethod
     def _build_get_profile_uri(
-            cls, profileId:str, schemaId:Optional[str], historic:bool=False, version:Optional[int]=None) -> str:
+            cls,
+            profileId: str,
+            schemaId: Optional[str],
+            historic: bool = False,
+            version: Optional[int] = None,
+            schemaless: Optional[bool] = None,
+            **kwargs) -> str:
         """
         Builds the url for retrieving a specific profile ...
         :param profileId:
@@ -69,20 +81,28 @@ class ProfilesRestClient(_Client):
         url_args = pydash.merge(
             {"historic": str(historic).lower()},  # Python turns True:bool into True:str
             {} if schemaId is None else {"schemaNames": schemaId},
-            {} if version is None else {"versionLimit": version}
+            {} if version is None else {"versionLimit": version},
+            {} if schemaless is None else {"schemaless": str(schemaless).lower()},
+            kwargs
         )
         return "{}?{}".format(cls.URIs["profile"].format(profileId=profileId), urllib.parse.urlencode(url_args))
 
-    def __init__(self, url, version, token):
-        super(ProfilesRestClient, self).__init__(url, version, token)
-
     def _done_finding_profiles(self, api_response:List, api_request_limit:Optional[int]):
         return (
            api_request_limit is not None and (len(api_response) < api_request_limit)
            or len(api_response) == 0
         )
 
-    def findProfiles(self, query:Optional[dict]=None, sort:Optional[dict]=None, limit:int=100, skip:int=0) -> List[ProfileSummary]:
+    def delete_cache_for_specific_profiles(self, profile_ids:List[str]) -> MessageResponse:
+        """
+        Invalidates cache for specific profiles ...
+        :param schema:
+        :return:
+        """
+        response = self._post_json(self.URIs["profile-cache"], {"profileIds": profile_ids})
+        return MessageResponse(**response)  # type: ignore
+
+    def findProfiles(self, query:Optional[dict]=None, sort:Optional[dict]=None, limit:int=100, skip:int=0, profileSchema:str=None) -> List[ProfileSummary]:
         """
         Finding profiles in the system.
 
@@ -94,6 +114,7 @@ class ProfilesRestClient(_Client):
         """
         uri = self.URIs["profiles"]
         uri_args = pydash.merge({},{},
+            {"schemaNames": profileSchema} if profileSchema is not None else {},
             {"filter": json.dumps(query)} if query is not None else {},
             {"sort": json.dumps(sort)} if sort is not None else {},
             {"limit": limit} if limit is not None else {},
@@ -162,7 +183,12 @@ class ProfilesRestClient(_Client):
             return False
         return True
 
-    def _get_profile(self, profileId:str, schemaId:Optional[str]=None, historic:bool=False, version:Optional[int]=None) -> Optional[List[dict]]:
+    def _get_profile(self,
+                     profileId: str,
+                     schemaId: Optional[str] = None,
+                     historic: bool = False,
+                     version: Optional[int] = None,
+                     **kwargs) -> Optional[List[dict]]:
         """
         Gets a profile using the cortex-graph api ...
 
@@ -173,8 +199,8 @@ class ProfilesRestClient(_Client):
         :return:
         """
 
-        uri = ProfilesRestClient._build_get_profile_uri(profileId, schemaId, historic=historic, version=version)
-        raw_profiles = self._get_json(uri) or {}
+        uri = ProfilesRestClient._build_get_profile_uri(profileId, schemaId, historic=historic, version=version, **kwargs)
+        raw_profiles = self._get_json(uri, debug=True) or {}
         return raw_profiles.get("profiles", [])
 
     def describeHistoricProfile(
@@ -199,7 +225,11 @@ class ProfilesRestClient(_Client):
         return p
 
     def describeProfile(
-            self, profileId:str, schemaId:str, version:Optional[int]=None) -> Optional[Profile]:
+            self,
+            profileId: str,
+            schemaId: Optional[str] = None,
+            version: Optional[int] = None,
+            **kwargs) -> Optional[Profile]:
         """
         Get the profile at a specific version ... only latest values of attributes are provided ...
 
@@ -209,9 +239,18 @@ class ProfilesRestClient(_Client):
         :return:
         """
         if schemaId is None:
-            raise Exception("schemaId required.")
+            warnings.warn("SchemaId was not provided; Attempting to retreive profile in a schemaless fashion.")
+            schemaless = "true"
+        else:
+            schemaless = "false"
         try:
-            profile = head(self._get_profile(profileId, schemaId, historic=False, version=version))
+            profile = head(
+                self._get_profile(
+                    profileId, schemaId,
+                    historic=False, version=version, schemaless=schemaless,
+                    **pydash.omit(kwargs, "schemaless")
+                )
+            )
             p = dict_to_attr_class(profile, Profile)
         except Exception as e:
             log.error(e)
@@ -220,7 +259,7 @@ class ProfilesRestClient(_Client):
         return p
 
     # Profile Ops ...
-    def deleteProfile(self, profileId: str, schemaId:Optional[str]=None) -> bool:
+    def deleteProfile(self, profileId: str, schemaId: Optional[str] = None) -> bool:
         """
         Deletes a profile ...
         :param profileId:
@@ -241,7 +280,7 @@ class ProfilesRestClient(_Client):
             return False
         return True
 
-    def pushEvents(self, events:List[EntityEvent]) -> List[str]:
+    def pushEvents(self, events: List[Union[EntityEvent, EntityRelationshipEvent]]) -> List[str]:
         """
         Pushes events to the profile service ...
         TODO ... make error collection better ...
@@ -262,7 +301,7 @@ class ProfilesRestClient(_Client):
         ]
 
     @deprecation.deprecated(deprecated_in='6.0.1b1', details='Use pushEvents instead.')
-    def pushAttributes(self, profileId:str, attributes:List[ProfileAttributeType]) -> List[str]:
+    def pushAttributes(self, profileId: str, attributes: List[ProfileAttributeType]) -> List[str]:
         """
         Pushes attributes to the latest profile for the specified profileId.
         Returns a list of messages with regards to the status of each attribute being pushed ...
@@ -307,7 +346,7 @@ class ProfilesRestClient(_Client):
 
     # PHASE 2 - Move the following to the server side!
 
-    def schemasForProfile(self, profileId:str) -> List[Optional[str]]:
+    def schemasForProfile(self, profileId: str) -> List[Optional[str]]:
         """
         Lists all of the schemas that the profile has been built against ...
 
@@ -319,11 +358,11 @@ class ProfilesRestClient(_Client):
         return [p.get("profileSchema") for p in profiles] if profiles else []
 
     def listVersions(self,
-                     profileId:str,
-                     schemaNameOrId:Optional[str]=None,
-                     limit:Optional[int]=None,
-                     after:Optional[str]=None,
-                     before:Optional[str]=None) -> List[ProfileVersionSummary]:
+                     profileId: str,
+                     schemaNameOrId: Optional[str] = None,
+                     limit: Optional[int] = None,
+                     after: Optional[str] = None,
+                     before: Optional[str] = None) -> List[ProfileVersionSummary]:
         """
         Lists all of the modifications done on the profile ... with the date of each modification!
         - [x] The logic for this has been moved to the cortex-graph service, it now has an api to list the different
